// Version 3: 
//   Data is allocated on heap, with pointers to it on the stack. 
//   AMPI virtualization ratios > 1 are supported.
//   AMPI migration is supported.

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>

#define BLOCKSIZE 2000

class chunk {
public:
  double t[BLOCKSIZE+2];    // working data array
  int xm, xp;        // xm: left processor number,  xp: right processor number
};

#ifdef AMPI
// p is a serialization operator which does packing operation on data when 
// an AMPI thread migrates out of a processor, or unpacking operation on data 
// when an AMPI thread migrates into a processor.
// void *d is the user data passed at time of calling MPI_Register.
// in this case, it is the chunk pointer (cp) 
void chunk_pup(pup_er p, void *d)
{
  chunk **cpp = (chunk **) d;
  if(pup_isUnpacking(p)) *cpp = new chunk;    // need to allocate chunk data when it is doing unpacking operation, i.e. when migrate into a processor
  chunk *cp = *cpp;
  pup_doubles(p, cp->t, (BLOCKSIZE+2));    // pup an array of doubles
  pup_int(p, &cp->xp);                     // pup an integer
  pup_int(p, &cp->xm);
  if(pup_isDeleting(p)) delete cp;         // free the memory when migrating out
}
#endif

#define abs(x) ((x)<0.0 ? -(x) : (x))

int main(int ac, char** av)
{
  int i, j, iter, niter;
  double error, tval, maxerr, tmpmaxerr;
  chunk *cp;
  int thisIndex, nblocks;
  double starttime, endtime, itertime;

  MPI_Init(&ac, &av);
  MPI_Comm_rank(MPI_COMM_WORLD, &thisIndex);
  MPI_Comm_size(MPI_COMM_WORLD, &nblocks);

  if (ac >= 2)
    niter = atoi(av[1]);
  else
    niter = 20;

  if (thisIndex == 0)
    printf("Running on %d MPI processors for %d iterations. \n", nblocks, niter);
  MPI_Bcast(&niter, 1, MPI_INT, 0, MPI_COMM_WORLD);

#if CMK_AIX
  cp = (chunk*)malloc(sizeof(chunk));
#else
  cp = new chunk;
#endif
#ifdef AMPI
    /* register the serialization function chunk_pup, the chunk pointer *cp is passed as user data. see chunk_pup function above */
  MPI_Register((void*)&cp, (MPI_PupFn) chunk_pup);
#endif

  cp->xp = (thisIndex+1)%nblocks;
  cp->xm = (thisIndex+nblocks-1)%nblocks;
  for(i=1; i<=BLOCKSIZE; i++)
    cp->t[i] = thisIndex*BLOCKSIZE+i-1;

  MPI_Barrier(MPI_COMM_WORLD);

  for(iter=1; iter<=niter; iter++) {
    starttime = MPI_Wtime();
    maxerr = 0.0;

    MPI_Request rreq[2];
    MPI_Status rsts[2];

    MPI_Irecv(&cp->t[0], 1, MPI_DOUBLE, cp->xm, 1, MPI_COMM_WORLD, &rreq[0]);
    MPI_Irecv(&cp->t[BLOCKSIZE+1], 1, MPI_DOUBLE, cp->xp, 0,
	      MPI_COMM_WORLD, &rreq[1]);

    MPI_Send(&cp->t[1], 1, MPI_DOUBLE, cp->xm, 0, MPI_COMM_WORLD);
    MPI_Send(&cp->t[BLOCKSIZE], 1, MPI_DOUBLE, cp->xp, 1, MPI_COMM_WORLD);

    MPI_Waitall(2, rreq, rsts);

    for (j=0; j<thisIndex*2+5; j++)            // some load imbalance
    for(i=1; i<=BLOCKSIZE; i++) {
      tval = (cp->t[i] + cp->t[i-1] + cp->t[i+1] ) / 3.0;
      error = abs(tval-cp->t[i]);
      cp->t[i] = tval;
      if (error > maxerr) maxerr = error;
    }
    MPI_Allreduce(&maxerr, &tmpmaxerr, 1, MPI_DOUBLE, MPI_MAX,
		  MPI_COMM_WORLD);
    maxerr = tmpmaxerr;

    endtime = MPI_Wtime();
    itertime = endtime - starttime;
    double  it;
    MPI_Allreduce(&itertime, &it, 1, MPI_DOUBLE, MPI_SUM,
                   MPI_COMM_WORLD);
    itertime = it/nblocks;

    if (thisIndex == 0) printf("iter %d time: %lf maxerr: %lf\n", iter, itertime, maxerr);
#ifdef AMPI
      // asking for load balancing for every 10 steps. This is a synchronous call (like a barrier)
    if(iter % 10 == 0) MPI_Migrate();
#endif
  }
  MPI_Finalize();
  return 0;
}
